#!/usr/bin/awk -f util.awk -f
#
# awkiawki - wikiwiki clone written in (n|g|m)awk
#
# Copyright (c) 2002 Oliver Tonnhofer <olt@bogosoft.com>
# Copyright (c) 2003-2008 Paul H. Liu <paul@thev.net>
#
# See the file `COPYING' for copyright notice.
#
# Here is a list of external programs it depends on:
#
#   system:   sort grep find md5sum mktemp chmod mv sed test mkdir rm test
#   revision: co ci rlog
#   image:    pbmtext math pnmcrop ppmtogif

BEGIN {
    # init ord chr
    _ord_init()
    # init the random
    srand()
    # init textdomain and locale support
    TEXTDOMAIN = "awki"
    bindtextdomain(".")

    SCRIPT = ENVIRON["SCRIPT_NAME"]
    SCRIPTDIR = dirname(SCRIPT)

    ###########################################################
    # below are system configurations
    # can be overidden by site.conf but must handle with care

    # command for parser
    # CMD_PARSER = "./" filename(SCRIPT)
    # command for computing md5sum, only returns the MD5 string
    CMD_MD5 = "md5sum|cut -d\\  -f1"
    # command for listing a single wiki file, with timestamp
    CMD_LIST_FILE = "find %s -printf '%p\\t%T@\\n'"
    # command for listing wiki files in DATADIR, sorted by time
    CMD_LIST_DIR = "find %s -name '_*_' -printf '%P\\t%T@\\n'|grep -v '/COMMENT/'"
    # command for listing all files in DATADIR, sorted by time
    CMD_LIST_DIR_ALL = "find %s -printf '%P\\t%T@\\n'|grep -v '/COMMENT/'"
    # command for sending (utf-8 encoded) email
    # arguments are: fromaddr toaddr subject
    CMD_SENDMAIL = "php mailit.php '%s' '%s' '%s'"
    # command for make a tempfile
    CMD_MKTEMP = "mktemp /tmp/temp.XXXXXX"
    # command for search, three arguments: datadir, grep option, keyword
    # CMD_SEARCH = "find %s -name '_*_' -exec grep -q %s '%s' {} ';' -printf '%P\\t%T@\\n'"
    # three arguments: grep option, keyword, file 
    CMD_SEARCH = "grep %s '%s' %s"
    # special_pages are not editable
    PAGE_SPECIAL = "SignIn|SignOut|ResetPass|TitleSearch|PageList|UserList|RecentChanges"
    # Regex for write protected files
    #   e.g.: "*", "PageOne|PageTwo|^.*NonEditable" 
    #   HINT: to edit these protected pages, upload a .htaccess 
    #         protected awki.cgi script with write_protection = ""
    LIMIT_WRITE = ""
    # the regexp for wiki names
    WIKIEXP = "[A-Za-z0-9:][A-Za-z0-9:. -]*"
    # the regexp for wiki page as in the URL
    WIKIEXPPAGE = "[A-Za-z0-9:][A-Za-z0-9:._-]*"
    # the regexp for wiki files in the filesystem
    WIKIEXPFILE = "_" WIKIEXPPAGE "_"
    # the regexp for wiki links in the page
    WIKIEXPLINK = "\\[(" WIKIEXP ")\\]"
    # the regexp for quoted links
    LINKEXP = "\\[([^|]+)\\|([^\\]]+)\\]"
    ###########################################################

    site_config()

    # init file operation extension if it's available
    # it's with awk's source distribution
    if (EXTENSION) {
      extension(EXTENSION, "dlload")
    }

    if (!ADMIN_EMAIL) {
       print "Content-type: text/html; charset=UTF-8\n\n" \
             "Please set a valid ADMIN_EMAIL in the AwkiAwki configuration file.\n"
       exit(0)
    }
    # initialization for default CGI
    # log this access in logfile 
    if (CMD_LOGFILE)
    {
        cmd = LOGFILE
        cmd | getline file
        close (cmd)
        print ENVIRON["REMOTE_ADDR"] " - - [" strftime("%D:%H:%M:%S %z") \
              "] \"" ENVIRON["REQUEST_METHOD"] " " ENVIRON["SCRIPT_NAME"] \
              ENVIRON["PATH_INFO"] "?" ENVIRON["QUERY_STRING"] "\" " \
              ENVIRON["REMOTE_HOST"] >> file
    }

    # get user info from cookie
    if (ENVIRON["HTTP_COOKIE"] && FILE_PASSWORD)
    {
        n = split(ENVIRON["HTTP_COOKIE"], cookies, ";")
        for (i = 1; i <= n; i++)
        {
            split(cookies[i], data, "=")
            cookie[gensub(/ /, "", 1, data[1])] = data[2]
        }
        cookie["user"] = clear_str(cookie["user"])
        cookie["hash"] = clear_str(cookie["hash"])
        cmd = "grep -i '^" cookie["user"] "\t' " FILE_PASSWORD "|cut -f2|tail -n1"
        cmd | getline md5
        close (cmd)
        cmd = "echo '" cookie["user"] ":" cookie["hash"] "'|" CMD_MD5
        cmd | getline hashmd5
        close (cmd)
        if (md5 && hashmd5 == md5)
        {
            AUTH_USER = cookie["user"]
        }
    }
    
    # PATH_INFO contains page name
    query["page"] = ENVIRON["PATH_INFO"]
    # query string
    query_str = ENVIRON["QUERY_STRING"]
    
    # process user upload via HTTP-POST
    if (ENVIRON["REQUEST_METHOD"] == "POST")
    {
        # normal form post
        if (ENVIRON["CONTENT_TYPE"] == "application/x-www-form-urlencoded")
        {
            # restrict the length to "max_post"
            if (ENVIRON["CONTENT_LENGTH"] < LIMIT_DATAPOST)
                bytes = ENVIRON["CONTENT_LENGTH"]
            else
                bytes = LIMIT_DATAPOST
            cmd = "dd ibs=1 count=" bytes " 2>/dev/null"
            # treat the form data as normal query string
            cmd | getline query_str1
            close (cmd)
            query_str = query_str1 "&" query_str
        }
        # file upload via MIME multipart
        else if (match(ENVIRON["CONTENT_TYPE"],
                 /multipart\/form-data; boundary=/) > 0)
        {
            boundary = substr(ENVIRON["CONTENT_TYPE"], RLENGTH + 1)
            parse_multipart(boundary)
            for (name in FORMDATA)
            {
                query[name] = FORMDATA[name]
            }
        }
    }
    
    # split the query_str into a table of key value pairs
    n = split(query_str, querys, "&")
    for (i = 1; i <= n; i++)
    {
        split(querys[i], data, "=")
        query[data[1]] = data[2]
    }

    # (IMPORTANT for security!)
    # TODO: we may need to check more stuff here
    query["page"] = clear_pagename(query["page"])
    query["backto"] = clear_pagename(query["backto"])
    query["revision"] = clear_revision(query["revision"])
    query["string"] = clear_str(urldecode(query["string"]))
    query["postid"] = urldecode(query["postid"])
    query["newname"] = clear_str(urldecode(query["newname"]))
    
    if (query["page"] == "")
        query["page"] = PAGE_DEFAULT

    if (query["page"] ~ PAGE_SPECIAL)
    {
        special_page = 1
    } 
    else
    {
        # the real file of the queried page in local filesystem
        query["pagefile"] = mk_pathname(query["page"])

        if (!LIMIT_WRITE || query["page"] !~ LIMIT_WRITE)
            page_editable = 1
    }
        if (query["random"])
        {
        print "Content-type: image/gif\n"
        srand(query["random"] * RANDOMKEY)
        math = sprintf("%04d", random(10000))
        system("pbmtext " math "| pnmcrop | ppmtogif")
        exit
    }
    if (query["postsave"])
    {
        # a little spammer blocker, to check the math result
        if (query["math"])
        {
            math = clear_str(urldecode(query["math"]))
            srand(math * RANDOMKEY)
            answer = revstr(clear_str(urldecode(query["answer"])))
            if (random(10000) != answer)
            {
                print "Content-type: text/html\n"
                print ""
                print "<html><body>"
                print _"REDOMATH"
                print "</body></html>"
                exit
            }
        }
    }
    else if (CMD=="rebuildmeta" || query["rebuildmeta"])
    {
        print "Content-type: text/html\n"
        print ""
        print "<html><body>Rebuild META file..."
        rebuild_meta(FILE_META)	  
        print "done.</body></html>"
        exit
    }

    # TODO: maybe should also make attachement redirect
    # we shall redirect the page immediately after an edit
    # to view its updated version
    if (query["save"] || query["postsave"] || query["postremoved"]) 
        redirect = SCRIPT "/" query["page"]

    if (query["pagefile"])
    {
        # get the file's last modified time without using rcs
        timestr = get_mtime(DATADIR query["pagefile"])
        date = sprintf(_"DATE_MODIFIED", sprintf(DATE_BOX, strftime(FMT_DATE, timestr)))
    }

    else if ((query["page"] == "SignIn") && query["user"])
    {
        # to verify login and redirect to the page it was from
        query["user"] = clear_str(query["user"])
        query["auto"] = clear_str(query["auto"])
        query["pass"] = clear_str(query["pass"])
        query["again"] = clear_str(query["again"])
        query["email"] = clear_str(query["email"])
        query["token"] = clear_str(query["token"])

        # only authenticate when token is not present
        # or token is present (this is the case of password reset)
        # together with pass and again.
        if (!query["token"] || (query["pass"] && query["again"]))
        {
            authmsg = authenticate(query["user"], query["auto"], 
                      query["pass"], query["again"], 
                      query["email"], query["token"])
            if (authmsg ~ /^ok/ && query["from"])
            {
                redirect = urldecode(query["from"])
            }
        }
    }
    else if (query["page"] == "SignOut")
    {
        AUTH_USER = ""
        authmsg = "logout"
        redirect = SCRIPT "/" PAGE_DEFAULT
    }

    header(query["page"], query["edit"], redirect, date, 
           authmsg, query["auto"], special_page)

    # process each command accordingly
    if (query["edit"] && page_editable)    
        edit(query["page"], query["pagefile"], query["revision"])
    else if (query["rename"] && page_editable)
        rename(query["page"], query["newname"])
    else if (query["save"] && query["text"] && page_editable)
        save(query["page"], query["text"], query["string"], 
             query["pagefile"])
    else if (query["post"])
        post(query["page"], query["postid"], query["quoteid"], query["quoteidx"])
    else if (query["postsave"] && query["text"])
        postsave(query["page"], query["postid"], query["text"])
    else if (query["postremove"] && query["postid"])
        postremove(query["page"], query["postid"])
    else if (query["postremoved"] && query["postid"])
        postremoved(query["page"], query["postid"])
    else if (query["postlist"])
        postlist(query["page"], query["pagefile"], query["postlist"])
    else if (query["upload"] && page_editable)
        upload(query["page"], query["uploadfile"], FORMFILE["uploadfile"], 
               FORMFILESIZE["uploadfile"], query["pagefile"], query["backto"])
    else if (query["page"] == "PageList")
        pagelist(DATADIR)
    else if (query["page"] == "UserList")
        userlist(DATADIR)
    else if (query["page"] == "RecentChanges")
        changes(DATADIR)
    else if (query["page"] == "TitleSearch")
        search(query["string"], query["case"], DATADIR)
    else if (query["page"] == "SignIn")
        login("", authmsg, query["user"], query["auto"], query["pass"], 
                  query["again"], query["email"], query["token"])
    else if (query["page"] == "ResetPass")
        reset(clear_str(query["user"]), clear_str(urldecode(query["email"])))
    else if (query["page"] && query["history"])
        history(query["page"], query["pagefile"])
    else if (query["diff"])
        diff(query["page"], query["pagefile"], 
            clear_str(query["r1"]), clear_str(query["r2"]))
    else if (query["revision"] || !redirect)
        parse(query["page"], query["pagefile"], query["revision"], 
              query["backto"])

    footer(query["page"], redirect)
}

# print header
function header(page, edit, redirect, date, authmsg, auto, special_page, title)
{

    title = page
    gsub("_", " ", title)
    # title = get_meta(mk_pathname(page)) " - " title
    visname = mk_linkedname(page)
    print "Content-type: text/html\n"
    print ""
    
    if (authmsg ~ /^ok/)
    {
        # set the user cookie
        hash = authmsg
        sub(/^ok:/, "", hash)
        if (auto == "on")
        {
            # when automatical signing in is turned on
            meta = "<meta http-equiv=\"Set-Cookie\" content=\"user=" \
                   AUTH_USER "; path=" SCRIPT "; expires=" \
                   strftime(FMT_DATE, systime() + 62208000) \
                   "\"><meta http-equiv=\"Set-Cookie\" content=\"hash=" hash \
                   "; path=" SCRIPT "; expires=" \
                   strftime(FMT_DATE, systime() + 62208000) \
                   "\" >"
        } 
        else
        {
            meta = "<meta http-equiv=\"Set-Cookie\" content=\"user=" \
                   AUTH_USER "; path=" SCRIPT "\" >" \
                   "<meta http-equiv=\"Set-Cookie\" content=\"hash=" hash \
                   "; path=" SCRIPT "\" >"
        }
    }
    else if (authmsg == "logout")
    {
        meta = "<meta http-equiv=\"Set-Cookie\" content=\"user=''; path=" \
                SCRIPT "\" >" \
               "<meta http-equiv=\"Set-Cookie\" content=\"hash=''; path=" \
                SCRIPT "\" >"
    }
    if (redirect)
    {   # all redirect page shall expire immediately
        print "<html><header>" meta \
	      "<meta http-equiv=\"Content-Type\" Content=\"text/html; charset=" \
              CHARSET "\"><meta http-equiv=\"expires\" content=\"" \
               strftime(FMT_DATE, systime() - 3600 * 48) \
               "\"> <meta http-equiv=\"Refresh\" content=\"1;URL=" \
               redirect "\" ></header>"
    }
    else
    {
	if (query["revision"]) editrev = "\\&revision=" query["revision"]
	else editrev = ""
        if (!special_page)
        {
          edt_href = "href"
          rev_href = "href"
        }
        else {
          edt_href = "name"
          rev_href = "name"
        }

        if (special_page || edit)
        {
            # some page shall expire immediately
            meta = meta "<meta http-equiv=\"expires\" content=\"" \
                   strftime(FMT_DATE, \
                            systime() - 3600 * 48) "\" >";
        }
        else 
        {
            meta = meta "<meta http-equiv=\"expires\" content=\"" \
                   strftime(FMT_DATE, \
                   systime() - 3600 * 48) "\" >";
        }
        if (FILE_HEADER)
        {
            while ((getline < FILE_HEADER) >0)
            {
                while (match($0, /[$]COOKIE:([^$]*)[$]/, r))
                {
                    $0 = substr($0, 1, RSTART - 1) clear_str(cookie[r[1]]) \
                    substr($0, RSTART + RLENGTH)
                }
                gsub(/[$]META[$]/, meta);
                gsub(/[$]TITLE[$]/, title);
                gsub(/[$]TITLELINK[$]/, visname);
                gsub(/[$]CHARSET[$]/, CHARSET)
                gsub(/[$]SEARCH[$]/, _"SEARCH")
                gsub(/[$]PAGE[$]/, page)
                gsub(/[$]SCRIPTNAME[$]/, SCRIPT)
                gsub(/[$]SCRIPTDIR[$]/, SCRIPTDIR)
                gsub(/[$]DEFAULTPAGE[$]/, PAGE_DEFAULT)
                gsub(/[$]EDITREV[$]/, editrev)
                gsub(/[$]EDT_HREF[$]/, edt_href)
                gsub(/[$]REV_HREF[$]/, rev_href)
                gsub(/[$]DATE[$]/, date)
                if (AUTH_USER) {
                    gsub(/[$]USERNAME[$]/, "[" mk_displayname(AUTH_USER) "]")
                    gsub(/[$]USERHOME[$]/, AUTH_USER)
                } else if (FILE_PASSWORD) {
                    gsub(/[$]USERNAME[$]/, _"SIGNIN")
                    gsub(/[$]USERHOME[$]/, "SignIn?from=" \
                         urlencode(SCRIPT "/" page))
                } else {
                    gsub(/[$]USERNAME[$]/, "")
                    gsub(/[$]USERHOME[$]/, "")
                }
                print
            }
            close(FILE_HEADER)
        }
    }
}

# print footer
function footer(page, redirect)
{
    if (redirect)
    {
        print "</html>"
    }
    else
    {
        if (FILE_FOOTER)
        {
            while ((getline < FILE_FOOTER) >0) print
            close(FILE_FOOTER)
        }
    }
}

function authenticate(user, auto, pass, again, email, token,    existing, file)
{
    if (email) email = urldecode(email);
    if (pass) pass = urldecode(pass);
    if (again) again = urldecode(again);

    if (!(user ~ /^[a-zA-Z]\w+$/)) {
        return "id_invalid"
    }
    # username must be a Wiki word
    user = clear_pagename(user)
    if (user == "") {
        return "id_invalid"
    }
    else if (length(user) < 3) {
        return "id_tooshort"
    }

    if (again)
    {
        # user registration
        file = DATADIR mk_pathname(user)
        if (!token && system("[ -f " file " ]") == 0 )
        {
            return "wikiname_exist"
        }
        cmd = "grep -i '^" user "\t' " FILE_PASSWORD "|cut -f2|tail -n1"
        cmd | getline existing
        close(cmd)
        if (pass != again)
        {
            return "pass_mismatch"
        }
        else if (pass ~ /\t\n\r`'"/)
        {
            return "pass_illegal"
        }
        else if (length(pass) < 5)
        {
            return "pass_tooshort"
        }
        else if (length(existing) > 0)
        {
            # already exist
            cmd = "grep -i '^" user "\t" token "\t' " FILE_RESET "|cut -f3|tail -n1"
            cmd |getline timelimit
            close(cmd)
            # check if it's password reset
            if (timelimit) {
                # check if the reset is timeout
                if (systime() > timelimit) {
                    return "timeout"
                }
                else {
                    printf("%s\t%s\t%s\n", user, token, systime() - 3600) >> FILE_RESET
                }
            }
            else return "id_exist"
        }
        # create this user
        cmd = "echo '" pass "'|" CMD_MD5
        cmd | getline hash
        close(cmd)
        cmd = "echo '" user ":" hash "'|" CMD_MD5
        cmd | getline hashmd5
        close(cmd)
        print user "\t" hashmd5 "\t" email >> FILE_PASSWORD
        fflush(FILE_PASSWORD)
        AUTH_USER = user
    }
    else
    {
        # user login
        if (pass ~ /\t\n\r`'"\\\//)
        {
            return "login_wrong"
        }
        cmd = "grep '^" user "' " FILE_PASSWORD "|cut -f2"
        cmd | getline md5
        close(cmd)
        cmd = "echo '" pass "'|" CMD_MD5
        cmd | getline hash
        close(cmd)
        cmd = "echo '" user ":" hash "'|" CMD_MD5
        cmd | getline hashmd5
        close(cmd)        
        if (md5 == hashmd5)
        {
            AUTH_USER = user
        }
    }
    if (AUTH_USER) 
    {
        return "ok:" hash
    }
    else
    {
        return "login_wrong"
    }
}

# function that handles login page and user info display
function login(from, authmsg, user, auto, pass, again, email, token)
{
    if (!from && query["from"]) from = urldecode(query["from"]);
    if (email) email = urldecode(email);
    if (pass) pass = urldecode(pass);
    if (again) again = urldecode(again);


    if (authmsg)
    {
        if (auto == "on") auto = "checked"
        if (authmsg ~ /^ok/)
        {
            print _"LOGIN_SUCCESS"
        }
        else if (authmsg == "id_invalid")
        {
            print _"ID_INVALID"
            user = ""
            userO = "<font color=\"red\">"
            userC = "</font>"
        }
        else if (authmsg == "id_tooshort")
        {
            print _"ID_TOOSHORT"
            user = ""
            userO = "<font color=\"red\">"
            userC = "</font>"
        }
        else if (authmsg == "wikiname_exist")
        {
            printf _"WIKI_EXIST", user
            user = ""
            userO = "<font color=\"red\">"
            userC = "</font>"
        } else if (authmsg == "id_exist")
        {
            printf _"ID_EXIST", user
            user = ""
            userO = "<font color=\"red\">"
            userC = "</font>"
        }
        else if (authmsg == "pass_mismatch")
        {
            print _"PASS_MISMATCH"
            pass = ""
            again = ""
            passO = "<font color=\"red\">"
            passC = "</font>"
        }
        else if (authmsg == "pass_illegal")
        {
            print _"PASS_ILLEGAL"
            pass = ""
            again = ""            
            passO = "<font color=\"red\">"
            passC = "</font>"
        }
        else if (authmsg == "pass_tooshort")
        {
            print _"PASS_TOOSHORT"
            pass = ""
            again = ""
            passO = "<font color=\"red\">"
            passC = "</font>"
        }
        else if (authmsg == "login_wrong")
        {
            print _"LOGIN_WRONG"
            pass = ""
            passO = "<font color=\"red\">"
            passC = "</font>"
        }
        else if (authmsg == "timeout")
        {
            printf(_"RESET_TIMEOUT", (SCRIPT "/ResetPass?user=" user))
            pass = ""
        }
        print "</font><p>"
    }
    if (AUTH_USER)
    {
        printf _"LOGGEDIN_PROMPT", " <a href=\"" SCRIPT "/" \
              AUTH_USER "\">" mk_displayname(AUTH_USER) "</a>"
        printf _"LOGIN_SIGNOUT", (SCRIPT "/SignOut")
        print "<br>"
    }
    else if (token)
    {
        if (authmsg != "timeout")
        {
            printf _"RESET_FORM", (SCRIPT "/SignIn"), user, \
                   passO, passC, pass, passO, passC, again, email, auto, token
        }
        return
    }
    else
    {
        print _"LOGIN_PROMPT"
    }

    if (!again)
    {
        printf "<br>"
        printf _"LOGIN_FORM", (SCRIPT "/SignIn"), \
	       userO, userC, user, passO, passC, auto, from
    }

    if (!AUTH_USER)
    {
        printf _"SIGNUP_FORM", (SCRIPT "/SignIn"), userO, userC, user, \
	       passO, passC, pass, passO, passC, again, email, auto, from
        printf _"RESETREQUEST_FORM", (SCRIPT "/ResetPass"), "", "", user, \
               "", "", email
        print _"SIGNUP_NOTICE"
    }
}

function reset(user, email,   cmd,userO,userC,emailO,emailC,email1,token,msg)
{
    cmd = "grep -i '^" user "\t' " FILE_PASSWORD "|cut -f3|tail -n1"
    cmd | getline email1
    if (email1 == "")
    {
        userO = "<font color=\"red\">" 
        userC = "</font>"
        printf cmd 
        printf _"RESET_NOSUCHUSER", user
    }
    else if (email)
    {
        if (email1 != email)
        {
            emailO = "<font color=\"red\">" 
            emailC = "</font>"
            printf _"RESET_EMAILNOTMATCH", user
        }
        else {
            token = random(10000000)
            printf("%s\t%s\t%s\n", user, token, systime() + 3600) >> FILE_RESET
            msg = sprintf(_"RESET_MSG", user, \
                    sprintf("http%s://%s:%d%s/SignIn?user=%s&email=%s&token=%s", \
                      ENVIRON["HTTPS"] ? "s" : "", \
                      ENVIRON["HTTP_HOST"], ENVIRON["SERVER_PORT"], \
                      SCRIPT, user, urlencode(email), token))
            sendmail(ADMIN_EMAIL, email, sprintf(_"RESET_SUBJECT", user), msg)
            printf(_"RESET_EMAILSENT", email)
            return
        }
    }
    printf _"RESETREQUEST_FORM", (SCRIPT "/ResetPass"), userO, userC, user, \
           emailO, emailC, email
} 

# send page to parser script
function parse(page, file, revision, backto, cmd)
{
    file = DATADIR file
    if (AUTH_USER && AUTH_USER == page)
    {
        printf _"LOGGEDIN_PROMPT", "<a href=\"" \
              SCRIPT "/" AUTH_USER "\">" mk_displayname(AUTH_USER) "</a>"
        printf _"LOGIN_SIGNOUT", (SCRIPT "/SignOut")
    }
    printf "<br>"

    # if the file really exists
    if (system("[ -f " file " ]") == 0)
    {
        #load_meta(FILE_META)

        if (revision)
        {
            printf _"DISPLAY_OLD", revision, (SCRIPT "/" page),
                   mk_displayname(page)
            cmd = "co -q -p'" revision "' " file
        }
        else cmd = "cat " file
        parse_wiki(cmd, page, file, revision)
    }
    else
    {
        # looks like an attachment?
        if (page ~ /^.*[.]([^.]+)$/)
        {
            # check if the file really exists after removing trailing _
            page1 = query["page"]
            sub(/_$/, "", page1)
	    file = DATADIR mk_pathname(page1)
            if (system("[ -f " file " ]") == 0)
            {
                printf _"UPLOAD_EXISTED", file #mk_displayname(page)
            }  
	    else {
            if (!AUTH_USER && FILE_PASSWORD)
            {
                #must sign in before upload
                printf _"SIGNIN_UPLOAD", (SCRIPT "/SignIn?from=" \
                       urlencode(SCRIPT "/" page "?attach=true&backto=" backto)) 
            }
            else
            {
                printf _"UPLOAD", (SCRIPT "?upload=true&page=" page), \
		       mk_displayname(page), query["backto"]
            }
            }
        }
        # otherwise, just plain edit
        else
        {
            printf _"BLANK", (SCRIPT "/" query["page"] "?edit=true")
        }
    }
}

# rename a page
function rename(page, newpage)
{
    if (!AUTH_USER && FILE_PASSWORD) {
        #must sign in before rename
        print _"SIGNIN_RENAME"
        login(SCRIPT "/" page "?edit=true")
        return
    }

    old_path = mk_pathname(page)
    if (newpage ~ /^:/)
    {
        sub(/^ *: */, "", newpage)
        gsub(" ", "_", newpage)
	newpage = km_pathname(dirname(old_path) "/_" newpage "_")
	gsub("_", " ", newpage)
    }
    
    new_path = mk_pathname(newpage)
    old_file = DATADIR old_path
    new_file = DATADIR new_path
    old_dir = DATADIR dirname(old_path) "/" shorter(filename(old_path))
    new_dir = DATADIR dirname(new_path) "/" shorter(filename(new_path))

    if (!match(newpage, "^" WIKIEXP "$") || newpage ~ /\./)
    {
        printf _"RENAME_INVALID", page, newpage
    }
    else if (system("test -f " new_file) == 0)
    {
        printf _"RENAME_EXIST", page, newpage
    }
    else if (system("test -f " old_file) != 0)
    {
        printf _"RENAME_NOTEXIST", page
    }
    else
    {
        system("mkdir -p '" DATADIR dirname(new_path) "'")
        system("mv '" old_file "' '" new_file "'")
        append_meta(FILE_META, "", old_path)
        #system("mv '" old_file ",v' '" new_file ",v'")
        system("mv '" old_dir  "' '" new_dir  "'")
        #print(old_path "," new_path "," old_dir  "," new_dir)
	append_meta(FILE_META, systime(), new_path)
	sub_dir = dirname(new_path) "/" shorter(filename(new_path))
	cmd = sprintf(CMD_LIST_DIR, new_dir)
        while ((cmd | getline fname) > 0)
	{
	    if (match(fname, /^(.*)\t(.*)$/, r))
	    {
		append_meta(FILE_META, r[2], sub_dir "/" r[1])
	    }
	}
	close(cmd)
        date = strftime(FMT_DATE, systime())
        
        if (AUTH_USER)
        {
            aopt = " -w" AUTH_USER
        }
        system("ci -q -t-" page aopt " -l -m'" ENVIRON["REMOTE_ADDR"] ";;" \
               date ";;renamed from " page "' " new_file)

        # replace old name occurrences in all pages
        cmd = "find '" DATADIR "' -name '_*_' -exec grep -q " page " {} \\; -print" 
        parse_rename(cmd, page, newpage, date, ENVIRON["REMOTE_ADDR"])
        close(cmd)
        printf _"RENAMED", page, (SCRIPT "/" km_pathname(new_path)), newpage
    }
}

# print edit form
function edit(page, file, revision, cmd)
{
    if (FILE_PASSWORD)
    {
        if (!AUTH_USER)
        {
            #must sign in before edit
            print _"SIGNIN_EDIT"
            login(SCRIPT "/" page "?edit=true")
            return
        }
        dir = gensub(":.*$", "", "", page)
        iscomment = match(page, ":COMMENT:")
        short = gensub("^.*:", "", "", page)
	if (dir)
        {
            cmd = "grep -i '^" dir "\t' " FILE_PASSWORD "|cut -f2"
            cmd | getline userdir
            close(cmd)
        }
	if (iscomment) 
        {
	    cmd = "grep -i '^" short "\t' " FILE_PASSWORD "|cut -f2"
            cmd | getline userfile
	    close(cmd)
        }
        # can't edit other's subpage unless its AUTH_USER's comment
	if (iscomment && ((userfile && AUTH_USER != short) || short ~ /anonymous[.]/))
        {
	    printf _"SIGNIN_EDIT_UNAUTH", short
            exit
        }
        if (userdir && AUTH_USER != dir && 
	    !(iscomment && userfile && AUTH_USER == short))
        {
  	    printf _"SIGNIN_EDIT_UNAUTH", dir
            exit
        }
    }
       
    # looks like an attachment?
    if (page ~ /^.*[.]([^.]+)$/)
    {
        # check if the file really exists after removing trailing _
        page1 = query["page"]
        sub(/_$/, "", page1)
        file = DATADIR mk_pathname(page1)
        if (system("[ -f " file " ]") == 0)
        {
            printf _"UPLOAD_EXISTED", file #mk_displayname(page)
        }  
        else {
        if (!AUTH_USER && FILE_PASSWORD)
        {
            #must sign in before upload
            printf _"SIGNIN_UPLOAD", (SCRIPT "/SignIn?from=" \
                   urlencode(SCRIPT "/" page "?attach=true&backto=" backto)) 
        }
        else
        {
            printf _"UPLOAD", (SCRIPT "?upload=true&page=" page), \
		   mk_displayname(page), query["backto"]
        }
        }
    }
    else
    {
        if (revision) printf _"EDIT_OVERWRITE", revision
        printf "<form action=\"" SCRIPT "?save=true&amp;page=" page \
              "\" enctype=\"multipart/form-data\" method=\"POST\">"
        printf "<br><br><textarea id=\"editbox\" name=\"text\" rows=25 cols=80>\n"

        # insert current page into textarea
        file = DATADIR file
        if (revision)
        {
            cmd = "co -q -p'"revision"' " file
            system(cmd)
        }
        else
        {
            while ((getline < file) >0)
                print
            close(file)
        }
        print "</textarea><br />"
        print _"EDIT_SAVE"
        print _"EDIT_SUMMARY"
        print "</form>"
        print _"HELP"
        printf _"EDIT_RENAME", (SCRIPT "?rename=true&amp;page=" page)
    }
}

# save page after a form submission from edit
function save(page, text, comment, file, dtext, date)
{
    if (!AUTH_USER && FILE_PASSWORD)
    {
        #must sign in before save
        print _"SIGNIN_SAVE"
        login(SCRIPT "/" page "?edit=true")
        return
    }

    dtext = text #urldecode(text)
    system("mkdir -p " dirname(DATADIR file))
    printf("%s", dtext) > DATADIR file
    close(DATADIR file)
    date = strftime(FMT_DATE, systime())
    append_meta(FILE_META, systime(), file)
    if (AUTH_USER)
    {
        aopt = " -w" AUTH_USER
    }
    system("ci -q -t-" page aopt " -l -m'"ENVIRON["REMOTE_ADDR"] ";;" date \
           ";;"comment"' " DATADIR file)
    printf _"SAVED", (SCRIPT "/" page), mk_displayname(page)
}


# print comment edit form
function post(page, postid, quoteid, quoteidx) {
    if (match(postid, /[:]/) > 1) {
        poster = substr(postid, RSTART + 1)
        if (poster != AUTH_USER) {
            printf _"SIGNIN_EDITAS", mk_displayname(poster)
            login(SCRIPT "/" page "?post=true")
            return
        }
    }

    if (!(AUTH_USER && FILE_PASSWORD)) {
        printf _"COMMENT_ANONYMOUS", ENVIRON["REMOTE_ADDR"], \
	       (SCRIPT "/SignIn?from=" urlencode(SCRIPT "/" page "?post=true"))
    } else {
        printf _"COMMENT_AS", (SCRIPT "/" AUTH_USER), mk_displayname(AUTH_USER)
    }
    print "<p><form action=\"" SCRIPT "?postsave=true&amp;page=" page \
          "\" method=\"POST\"></p>"

    # this is a trick not to leave a newline at the beginning of the
    # form text area
    old_ORS = ORS
    ORS = ""
    print "<textarea name=\"text\" rows=12 cols=80>"
    ORS = old_ORS

    if (postid) {
        file = DATADIR mk_pathname(page ":COMMENT:" postid)
        if (system("[ -f " file " ]") == 0 ) {
            while ((getline < file) >0)
                print
            close(file)
        }
    }

    if (match(quoteid, /^(.*):(.*)$/, r))
    {
        file = DATADIR mk_pathname(page ":COMMENT:" quoteid)
        if (system("[ -f " file " ]") == 0 ) {
	    printf _"POST_INREPLY", quoteidx, r[2]
	    print "\n<blockquote>"
            while ((getline < file) >0) print
            close(file)
	    print "</blockquote>"
	} 
    }

    print "</textarea><br />"

    if (!(AUTH_USER && FILE_PASSWORD)) {
      math = random(1000)
      printf _"SPAMMER", (SCRIPT "?random=" math), math
    }
    print "<input type=\"submit\" value=\"save\">"
    print "<input type=\"hidden\" name=\"postid\" value=\"" postid "\">"
    print "</form>"
    print _"HELP"
}

# post comment after a form submission from comment
function postsave(page, postid, text) {

    if (match(postid, ":") > 0) {
        poster = substr(postid, RSTART + 1)
        if (poster != AUTH_USER) {
            printf _"SIGNIN_POSTCOMMENT", mk_displayname(poster)
            login(SCRIPT "/" page "?post=true")
            return
        }
    } else {
        if (!(AUTH_USER && FILE_PASSWORD)) {
            postid = systime() ":anonymous." ENVIRON["REMOTE_ADDR"]
        } else {
            postid = systime() ":" AUTH_USER
        }
    }
    file = mk_pathname(page ":COMMENT:" postid)
    system("mkdir -m 755 -p '" dirname(DATADIR file) "'")
    dtext = urldecode(text);
    printf dtext > DATADIR file
    close(DATADIR file)
    append_meta(FILE_META, systime(), gensub(/^(.*\/COMMENT)\/.*$/, "\\1", "g", file))
    date = strftime(FMT_DATE, systime())
    if (AUTH_USER) {
        aopt = " -w" AUTH_USER
    }

    system("ci -q -t-" page aopt " -l -m'" ENVIRON["REMOTE_ADDR"] ";;" date \
           ";;' " DATADIR file)

    printf _"POSTED", (SCRIPT "/" page), mk_displayname(page)
}

# remove a comment form
function postremove(page, postid) {
    if (!AUTH_USER && FILE_PASSWORD) {
        #must sign in before removing comment
        print _"SIGNIN_REMOVE"
        login(SCRIPT "/" page "?postremove=true&postid=" postid)
        return
    }

    if (match(postid, ":") > 0) {
        posttime = substr(postid, 1, RSTART - 1)
        poster = substr(postid, RSTART + 1)
        if (poster != AUTH_USER) {
            printf _"REMOVE_AS", mk_displayname(poster)
            login(SCRIPT "/" page "?postremove=true&postid=" postid)
            return
        }
    }

    printf _"COMMENT_BY", mk_displayname(poster), strftime(FMT_DATE, posttime)

    file = DATADIR mk_pathname(page ":COMMENT:" postid)

    if (system("[ -f " file " ]") == 0 ) {
        while ((getline < file) >0)
            print
        close(file)
    }

    printf _"COMMENT_REMOVE", (SCRIPT "/" page "?postremoved=true"), postid
}

# remove a comment form
function postremoved(page, postid) {
    if (!AUTH_USER && FILE_PASSWORD) {
        #must sign in before removing comment
        print _"SIGNIN_REMOVE"
        login(SCRIPT "/" page "?postremove=true&postid=" postid)
        return
    }

    if (match(postid, /:/) > 1) {
        posttime = substr(postid, 1, RSTART - 1)
        poster = substr(postid, RSTART + 1)
        if (poster != AUTH_USER) {
            printf _"REMOVE_AS", mk_displayname(poster)
            login(SCRIPT "/" page "?postremove=true&postid=" postid)
            return
        }
    }

    file = mk_pathname(page ":COMMENT:" postid)
    printf "" > DATADIR file
    close(DATADIR file)
    append_meta(FILE_META, systime(), gensub(/^(.*)\/COMMENT\/.*$/, "\\1", "g", file))
    print _"REMOVED"
}

function postlist(name, file, pageno) {
    #load_meta(FILE_META)
    # if the file really exists
    file = DATADIR file
    if (system("[ -f " file " ]") == 0 ) {
        index_anchor = 0
        index_level = 0
        list["ol"] = 0
        list["ul"] = 0
        indent_level = 0
        blankline = 1
        FS = "[ ]"
	title = -1

        cmd = "grep '^<comment' " file 
        parse_wiki(cmd, name, file, "", 0, pageno)
    } else {
        printf _"BLANK", (SCRIPT "/" query["page"] "?edit=true")
    }
}

# upload file

function upload(page, upload_file, tempfile, size, file, existed, backto) {

    if (!AUTH_USER && FILE_PASSWORD)
    {
        #must sign in before upload
        print _"SIGNIN_UPLOAD1"
        login(SCRIPT "/" page "?attach=true")
        return
    }

    if (page ~ /^.*[.]([^.]+)$/)
    {
	file = dirname(file) "/" shorter(filename(file))
	sub(/^\//, "", file)
        upload_file = DATADIR file
    }
    else
    {
        print _"UPLOAD_FILENAME"
        return
    }

    if ((file == "") || (size == 0))
    {
        print _"UPLOAD_INVALID"
    }
    else
    {
        cmd = "test -f '" DATADIR upload_file "' && echo 1"
        cmd | getline existed
        close(cmd)
      
        if (existed == 1)
        {
            printf _"UPLOAD_EXISTED", upload_file
        }
        else
        {
            system("mkdir -m 755 -p '" dirname(upload_file) "'")
            system("chmod go+r " tempfile)
            system("mv '" tempfile "' '" upload_file "'")
	    if (tolower(upload_file) ~/(jpg|gif|png|tiff|tif)/) {
		system("./fitimg.sh '" upload_file "' 2> fitimg.out")
		if (system("test -e '" upload_file "_'") == 0)
                  append_meta(FILE_META, systime(), file "_")
            }
            append_meta(FILE_META, systime(), file)
            printf _"UPLOADED", mk_displayname(page)
        }
    }
    # after the upload, just go back to original page
    printf _"UPLOAD_BACK", (SCRIPT "/" query["backto"]),
           mk_displayname(clear_pagename(query["backto"]))
           
}

# TODO: recurse into dir
# list all pages (sorted by name)
function pagelist(datadir,    page, names, pages, n, i, t)
{
    load_meta(FILE_META)
    n = 0
    for (name in META)
    {
        if (name ~ /_$/ && META[name])
        {
            page = km_pathname(name)
            if (page != "") 
            {
                pages[++n] = page
                names[page] = name
            }
        }
    }
    asort(pages)
    print _"PAGELIST" "<table width='100%'>" _"PAGELIST_TH"
    for (i = 1; i <= n; i++)
    {
        page = pages[i]
        name = names[page]
	t = gensub("_", " ", "g", page)
        printf _"PAGELIST_TR", mk_simple_link(page, t), \
	       mk_simple_link(page, META[name ":"])
    }
    print "</table>"
}

# list all users (sorted by name)
function userlist(datadir,    page, pages, n, i, p, t, last)
{
    n = 0
    while (getline l < FILE_PASSWORD)
    {
	match(l, /^([^\t]*)\t.*$/, r);
	if (t = get_meta("_" r[1] "_")) {
		pages[++n] = t "|" r[1]
	}
    }
    close(FILE_PASSWORD)
    asort(pages)
    print _"USERLIST" "<table width='100%'>" _"USERLIST_TH"
    for (i = n; i >= 1; i--)
    {
        match(pages[i], /^(.*)\|(.*)$/, r);
	page = r[2]
	t = r[1]
	if (length(page) > 0 && page != last) {
          p = gensub("_", " ", "g", page)
          t = strftime(FMT_DATE, t)
          printf _"USERLIST_TR", mk_simple_link(page, p), \
                 mk_simple_link(page, get_meta("_" page "_:")), \
		 sprintf(DATE_BOX, t)
	}
        last = page
    }
    print "</table>"
}

# TODO: recurse into dir
# list pages with last modified time (sorted by date)
function changes(datadir, date,   
    pages, m, n, i, page, rev, name, yesterday, link)
{
    load_meta(FILE_META)
    n = 0
    for (name in META)
    {
        if (name ~ /(_|COMMENT)$/ && META[name])
        {
	    pages[++n] = META[name] "|" name
	}
    }
    n = asort(pages)
    date = strftime(FMT_DATE, systime())
    printf _"DATE_NOW", (sprintf(DATE_BOX, date))
    print _"RECENT_CHANGE"
    i = n
    m = 0
    print "<table width='100%'>" _"RECENT_TH"
    yesterday = systime() - 24 * 3600
    while (i > 0)
    {
	if (!(pages[i] ~ /\/COMMENT$/) && (pages[i] ~ /(\||\/)_(.*)_$/) &&
            match(pages[i], /^(.*)\|(.*)$/, r))
	{
	    page = km_pathname(r[2])
	    name = gensub("_", " ", "g", page)
	    rev = r[1]
	    t = strftime(FMT_DATE, rev)
	    printf _"RECENT_TR", mk_simple_link(page, name), \
		   mk_simple_link(page, META[r[2] ":"]), \
		   sprintf(DATE_BOX, t)
	    if (m++ > LIMIT_CHANGES && rev < yesterday) break;
	}
	i--
    }
    print "</table><br>"
    print _"RECENT_COMMENT"
    i = n
    m = 0
    print "<table width='100%'>" _"RECENT_TH"
    while (i > 0)
    {
	if (match(pages[i], /^(.*)\|(.*)\/COMMENT/, r))
	{
	    link = r[2]
            link = gensub(/(^|\/)([^/]*)$/, "\\1_\\2_", "g", link)
	    page = km_pathname(link)
	    name = gensub("_", " ", "g", page)
	    rev = r[1]
	    t = strftime(FMT_DATE, rev)
	    printf _"RECENT_TR", mk_simple_link(page, name), \
		   mk_simple_link(page, META[link ":"]), \
		   sprintf(DATE_BOX, t)
	    if (m++ > LIMIT_CHANGES && rev < yesterday) break;
	}
	i--
    }
    print "</table>"

}

# keyword search
function search(name, case, datadir,
    cmd, opt)
{
    #load_meta(FILE_META)
    if (case ~ /sensitive/) { opt = "" }
    else { opt = "-i" }
    printf _"SEARCH_RESULT", name
    #cmd = sprintf(CMD_SEARCH, DATADIR, opt, name)
    cmd = sprintf(CMD_SEARCH, opt, name, FILE_META)
    parse_search(cmd)
    print name "\t" systime() >> FILE_SEARCH
    close(FILE_SEARCH)
}

# revision history of a page
function history(page, file)
{
    printf _"HISTORY_LIST", (SCRIPT "/" page), mk_displayname(page)
    RS = "----------------------------\n"
    FS = "\n"
    print "<table border=1>\n" _"HISTORY"
    num_revisions = 0
    cmd = "rlog " DATADIR file
    parse_history(cmd, page)
    print "</table>"
    printf _"DIFF_FORM", SCRIPT "/" page
    printf "<input type='hidden' name='diff' value='true'>"
    print "<select name='r1'>"
    n = 0
    while (n++ < num_revisions)
    {
        printf "<option value='%s'>%s</option>", revisions[n],  revisions[n]
    }
    print "</select>&nbsp;"
    print "<select name='r2'>"
    n = 0
    while (n++ < num_revisions)
    {
        printf "<option value='%s'>%s</option>", revisions[n],  revisions[n]
    }
    print "</select>&nbsp;" _"DIFF_BUTTON" "</form>"
}

# display a diff of two revisions
function diff(page, file, r1, r2)
{
    file = DATADIR file
    # if the file really exists
    if (system("[ -f " file " ]") == 0)
    {
        printf _"DIFF", r1, r2
        system("rcsdiff -r'" r1 "' -r'" r2 "' -u -q --strip-trailing-cr " file " | " \
               "sed -e 's/</\\&lt;/g' -e 's/>/\\&gt;/g' " \
               "-e 's/  / \\&nbsp;/g' "\
               "-e 's/^+\\(.*\\)$/<div style=\"width: 100%;background-color: #d0ffd0;\">\\1\\&nbsp;<\\/div>/' " \
               "-e 's/^@@\\(.*\\)@@$/<h2 style=\"color: #ff8020;\">\\1<\\/h2>/' " \
               "-e 's/^-\\(.*\\)$/<div style=\"width: 100%;background-color: #ffd0d0;\">\\1\\&nbsp;<\\/div>/' " \
               "-e 's/^ \\(.*\\)$/<div style=\"width: 100%;\">\\1\\&nbsp;<\\/div>/g' " \
              )
    }
}



# remove '"`\/ characters from string
# and limit size to 80 characters
# *** !Important for Security! ***
function clear_str(str)
{
    gsub(/['`"\/\\]/, "", str)
    if (length(str) > 80)
        return substr(str, 1, 80)
    else
        return str
}

# clean up pagename string
# *** !Important for Security! ***
function clear_pagename(str)
{
    if (match(str, WIKIEXPPAGE) > 0)
        return substr(str, RSTART, RLENGTH)
    else
        return ""
}

# clean up revision numbers
# *** !Important for Security! ***
function clear_revision(str)
{
    if (match(str, /[1-9]\.[0-9]+/))
        return substr(str, RSTART, RLENGTH)
    else
        return ""
}

function shorter(link)
{
    sub(/^_/, "", link)
    sub(/_$/, "", link)
    return link
}

# decode urlencoded string
function urldecode(text, hex, i, hextab, decoded, len, c, c1, c2, code)
{
    split("0 1 2 3 4 5 6 7 8 9 a b c d e f", hex, " ")
    for (i=0; i<16; i++) hextab[hex[i+1]] = i

    # urldecode function from Heiner Steven
    # http://www.shelldorado.com/scripts/cmds/urldecode

    # decode %xx to ASCII char 
    decoded = ""
    i = 1
    len = length(text)
    
    while ( i <= len )
    {
        c = substr (text, i, 1)
        if ( c == "%" )
        {
            if ( i+2 <= len )
            {
                c1 = tolower(substr(text, i+1, 1))
                c2 = tolower(substr(text, i+2, 1))
                if ( hextab [c1] != "" || hextab [c2] != "" )
                {
                    if ( (c1 >= 2 && !(c1 == 7 && c2 == "F")) || 
                         (c1 == 0 && c2 ~ "[9acd]") )
                    {
                        code = 0 + hextab [c1] * 16 + hextab [c2] + 0
                        c = sprintf ("%c", code)
                    } 
                    else
                    {
                        c = " "
                    }
                    i = i + 2
               }
            }
        }
        else if ( c == "+" )
        {    # special handling: "+" means " "
            c = " "
        }
        decoded = decoded c
        ++i
    }
    
    # change linebreaks to \n
    gsub(/\r\n/, "\n", decoded)
    
    # remove last linebreak
    sub(/[\n\r]*$/,"",decoded)
    
    return decoded
}

function urlencode(string,   nord)
{
# We assume an awk implementation that is just plain dumb.
# We will convert an character to its ASCII value with the
# table ord[], and produce two-digit hexadecimal output
# without the printf("%02X") feature.

    EOL = "%0A"                # "end of line" string (encoded)
    split ("1 2 3 4 5 6 7 8 9 A B C D E F", hextab, " ")
    hextab [0] = 0
    for (i = 1; i <= 255; ++i ) nord [ sprintf ("%c", i) "" ] = i + 0

    encoded = ""
    for (i = 1; i <= length (string); ++i)
    {
        c = substr (string, i, 1)
        if ( c ~ /[a-zA-Z0-9.-]/ )
        {
            encoded = encoded c# safe character
        }
        else if ( c == " " )
        {
            encoded = encoded "+"# special handling
        }
        else
        {
            # unsafe character, encode it as a two-digit hex-number
            lo = nord [c] % 16
            hi = int (nord [c] / 16);
            encoded = encoded "%" hextab [hi] hextab [lo]
        }
    }
    return encoded
}


# load configfile
function load_config(script, configfile, key, value, extended)
{
    configfile = script
    #remove trailing / ('/awki/awki.cgi/' -> '/awki/awki.cgi')
    sub(/\/$/, "", configfile)
    #remove path ('/awki/awki.cgi' -> 'awki.cgi')
    sub(/^.*\//, "", configfile)
    #remove suffix ('awki.cgi' -> 'awki')
    sub(/\.[^.]*$/,"", configfile)
    # append .conf suffix
    configfile = configfile ".conf"
    extended = 0
    #read configfile
    while((getline < configfile) >0)
    {
        if (extended && key) {
           sub(/[ \t]*$/, "")
           sub(/^[ \t]*/, "")
           if (sub(/\\$/, "")) extended = 1
           else extended = 0
           localconf[key] = localconf[key] $0
        }
        else if ($0 ~ /^#/) continue #ignore comments        
        else if (match($0,/[ \t]*=[ \t]*/))
        {
            key = substr($0, 1, RSTART-1)
            sub(/^[ \t]*/, "", key)
            value = substr($0, RSTART+RLENGTH)
            if (sub(/\\$/, "", value)) extended = 1
            if (sub(/^"/, "", value))
                sub(/"$/, "", value) 
            #set conf variables
            localconf[key] = value
        }
    }
    close(configfile)
}    

# lookup the localconf for given name, return default value if not found.
function getconf(name, default)
{
    return (localconf[name] ? localconf[name] : default)
}

# standard parser functions go below
function match_level(spaces,    i)
{
#printf "in match_level = %d spaces = %d<br>", indent_level, spaces
    for (i = 1; i <= indent_level; i++)
    {
        if (indent_levels[i] >= spaces) break
    }
    if (i > indent_level)
    {
        indent_levels[++indent_level] = spaces
    }
    else {
        if (i < indent_level) close_tags(indent_level - i)
        return 1
    }
    return 0
#printf "out match_level = %d i = %d indent_tag = %s<br>", indent_level, i, indent_tags[indent_level]
}

function open_tags(type, item, desc)
{
    if (type == "any") {
	type = indent_tags[indent_level] ? indent_tags[indent_level] : "ul";
    }
    if (indent_tags[indent_level] != type)
    {
        if (indent_tags[indent_level]) { close_tags(1); indent_level++ }
        indent_tags[indent_level] = type
        printf OPENTAG[type] #, indent_levels[indent_level] * 5
    }
    if (desc) print "<dt>" desc
    if (item) print item
}

function close_tags(n)
{
#printf "in close_tags n = %d indent_level = %d<br>", n, indent_level
    if (!n) n = indent_level
    while (n && indent_level)
    {
        printf CLOSETAG[indent_tags[indent_level]], n--
        indent_tags[indent_level--] = ""
    }
#printf "out close_tags indent_level = %d<br>", indent_level
}

function next_anchor() {
    index_anchor = index_anchor + 1
    return ("<a name=\"anchor_" (index_anchor - 1) "\"></a>")
}

function mk_link(page, link, display, deepquote,   r, e, path, name, file, ext, rev) {
    RAWLINK=""
    if (match(link, "^" WIKIEXP "$")) {
	gsub(/ *: */, ":", link)
        gsub(/ /, "_", link)
        if (match(link, /^(.*):([^:]+)$/, r)) {
            path = r[1] ":"
            name = r[2]
            if (path ~ /^:/) path = shorter(page) path
            # remove all .. and . in path
            gsub(/\./, "", path)
            # remake the clean link
            link = path name
            # convert path to directory
            gsub(/:/, "/", path)
        }
        else {
            path = ""
            name = shorter(link)
        }
        file = "_" name "_"
        rev = get_meta(path file)
	if (!display && deepquote) display = get_meta(path file ":")
        sub(/^( |　)+/, "", display)
        sub(/( |　)+$/, "", display)
        # handle direct file link (only those with file extension)
        if (get_meta(path name) && match(name, /^.*[.]([^.]+)$/, r)) {
            # handle file with extension differently
            ext = tolower(r[1])
            RAWLINK = DATAURL path name

            if (deepquote && (ext ~ /jpg|gif|png|tif/)) {
                # embed image
		if (get_meta(path name "_")) {
		   return "<a href=\"" RAWLINK "\"><img border=\"0\" src=\"" \
                          DATAURL path name "_\"></a>"
                }
                else return "<img src=\"" RAWLINK "\">" 
            }
            else {
                gsub(/_/, " ", name)
                if (display) 
                    return "<a href=\"" RAWLINK "\">" display "</a>" 
                return "<a href=\"" RAWLINK "\">" name "</a>"
            }
        }
        gsub(/_/, " ", name)
        if (rev) {
          RAWLINK = SCRIPT "/" link
          if (display)
	    return "<a href=\"" RAWLINK "\">" display "</a>"
          else
	    return "<a href=\"" RAWLINK "\">" name "</a>"
        }
        else {
          RAWLINK = SCRIPT "/" link "?edit=true&backto=" page
          if (display) 
              return display "<a href=\"" RAWLINK "\">?</a>" 
          return name "<a href=\"" RAWLINK "\">?</a>" 
        }
    }
    else if (link ~ /^\[?((https?|ftp|gopher):\/\/|(mailto|news):)[^\]\t]*\]?/) {
	sub(/^\[/, "", link)
	sub(/\]$/, "", link)
        RAWLINK=link
        e = "((https?|ftp|gopher):\\/\\/|(mailto|news):)[^\\t]*[^.,?;:\\'\")\t]"
        #sub(e, "&", RAWLINK)
        # remove mailto: in link description
        if (tolower(RAWLINK) ~ /(jpg|gif|png|tif)$/) {
            sub(e, "<img src=\"&\">", link)
        }
        else {
            sub(e, "<a href=\"&\">&</a>", link)
            sub(/>mailto:/, ">", link)
        }
        if (display)
            return "<a href=\"" RAWLINK "\">" display "</a>"
        return link
    }
    else
    {
        return link
    }
}

# make a simple link from name and revision
function mk_simple_link(name, display) {
    if (name ~ WIKIEXP) {
        if (!display) display = mk_displayname(name)
        return "<a href=\"" SCRIPT "/" name "\">" short_title(display) "</a>"
    }
    else return name
}

function close_ul(l)
{
    while (index_level > l) { 
    print("</ul>"); index_level = index_level - 1;
    }
}

function open_ul(l)
{
    while (index_level < l) { 
    print("<ul>"); index_level = index_level + 1;
    }
}


# convert a wiki name into a display name with every
# part of the parent pages linked
function mk_linkedname(name,     i, s, t, n, u)
{
    s = ""
    i = 0
    n = name ":"
    while (match(n, ":") > 1) 
    {
        t = substr(n, 1, RSTART - 1)
        gsub(/_/, " ", t)
        u = sprintf("<a href=\"%s\">%s</a>", \
            SCRIPT "/" substr(name, 1, i + RSTART - 1) , t)
	s = s ? s " : " u : u
        n = substr(n, RSTART + 1)
	i = i + RSTART
    }
    gsub(/_/, " ", n)
    return s
}

function parse_multipart(boundary,   s, r, name, filename, tempfile) {
    while (getline s)
    {
        RS = "\n"
	name = ""
        filename = ""
	tempfile = ""
        if (match(s, /^Content-Disposition:.* name="([^"]*)"/, r)) name = r[1]
        if (match(s, /^Content-Disposition:.* filename="([^"]*)"/, r)) 
	{
	    filename = r[1]
	    cmd = CMD_MKTEMP
            cmd | getline tempfile
            close (cmd)
	}
	if (!name) continue
        while (getline s) { if (s == "" || s == "\r") break }
        BINMODE = "rw"
        gsub(/\+/, "\\+", boundary)
        RS = "\r\n--" boundary
        getline s
        ORS = ""
        OFS = ""
        if (tempfile) {
          print s > tempfile
          close(tempfile)
          FORMDATA[name] = filename
	  FORMFILE[name] = tempfile
	  FORMFILESIZE[name] = length(s)
	}
        else {
	  FORMDATA[name] = s
	}
        length($0)
        RS = "\r\n"
	getline s
    }
    RS="\n"
    ORS = "\n"
    OFS = " "
}
