# Cache support (poorman's database)
# 
# It's to cache the page title and timestamp of all files for faster access.
# It's not necessary to load_meta(..) every time, and in the case where META
# is not loaded, get_meta(..) will do real time lookup. 
# 
# NOTE: Loading a few hundred kilobytes of META is still faster than spawning 
#       a process to check a file's timestamp.
# 
# The META file should always be kept up to date, i.e., whenver a file is 
# updated, its info should go into META.
# 
# The META is organized as key/value pairs:
#
#    1. key -> actual file name (without the DATADIR prefix)
#       value -> timestamp
#    2. key -> actual file name ":"
#       value -> its page title (when it's really a page)
#
# It can be expanded in the future to support more key and value types.
#
# Global Used: 
#   DATADIR, META, TEMP_META, CMD_LIST_FILE, CMD_LIST_DIR_ALL, LINKEXP, WIKIEXPLINK,
#   WIKIEXPPAGE, $0

# load meta file
function load_meta(metafile, 
    key, value, extended)
{
    while((getline < metafile) >0)
    {
	sub(/\r$/, "")
        if (extended && key)
        {
           sub(/[ \t]*$/, "")
           sub(/^[ \t]*/, "")
           if (sub(/\\$/, "")) extended = 1
           else extended = 0
           META[key] = META[key] $0
        }
        else if ($0 ~ /^#/) continue #ignore comments        
        else if (match($0,/[ \t]*=[ \t]*/))
        {
            key = substr($0, 1, RSTART-1)
            sub(/^[ \t]*/, "", key)
            value = substr($0, RSTART+RLENGTH)
            if (sub(/\\$/, "", value)) extended = 1
            if (sub(/^"/, "", value))
                sub(/"$/, "", value) 
            #set conf variables
            META[key] = value
        }
    }
    META["loaded"] = 1
    close(metafile)
}    

# get meta value for given key
function get_meta(key,
    file, cmd, fname, filetime, line, ret, timestr)
{
    if (!META["loaded"]) 			# META not loaded
    {
	if (TEMP_META[key]) return TEMP_META[key]
	file = key
	sub(/:$/, "", file)
	fname = DATADIR file
	mtime = get_mtime(fname)
        if (mtime)
        {
            TEMP_META[file] = mtime
	    if ((fname ~ (WIKIEXPFILE "$")) && 
		(getline line < fname) > 0)
            {		
                sub(/^-+/, "", line)
                sub(/\r/, "", line)
	        sub(/^ +/, "", line)
	        sub(/ +$/, "", line)
	        line = gensub(LINKEXP, "\\2", "g", line)
	        line = gensub(WIKIEXPLINK, "\\1", "g", line)
		TEMP_META[file ":"] = line
	    }
            close(fname)
	    return TEMP_META[key]
        }
        else return NULL
    }
    else return META[key]
}

# rebuild meta file
function rebuild_meta(metafile, 
    cmd, ls_out, page, rev, fname, n)
{
    print "" > metafile
    cmd = sprintf(CMD_LIST_DIR_ALL, DATADIR)

    while ((cmd | getline ls_out) >0)
    {
        if (match(ls_out, "^([^\t]+)(\t[0-9a-fA-F]+)?$", r))
        {
            page = r[1]
            rev = gensub(/\t/, "", 1, r[2])
            if (!rev) rev = 1
            append_meta(metafile, rev, page)
        }
    }
    close(cmd)
    close(metafile)
}

# append updated meta to the end of meta file
function append_meta(metafile, timestamp, file,     
    fname, n)
{
    printf "%s = %s\n", file, timestamp >> metafile
    fname = DATADIR file
    if ((getline < fname) > 0)
    {
	sub(/\r/, "")
        sub(/^-+/, "")
	sub(/^ +/, "")
	sub(/ +$/, "")
	$0 = gensub(LINKEXP, "\\2", "g", $0)
	$0 = gensub(WIKIEXPLINK, "\\1", "g", $0)
	printf "%s = %s\n", (file ":"), $0 >> metafile
    }
    if ((getline < fname) > 0 && ($0 ~ /^ /))
    {
	sub(/\r/, "")
        sub(/^-+/, "")
	sub(/^ +/, "")
	sub(/ +$/, "")
	$0 = gensub(LINKEXP, "\\2", "g", $0)
	$0 = gensub(WIKIEXPLINK, "\\1", "g", $0)
	if ($0 != "") 
  	    printf "%s = %s\n", (file ":1"), $0 >> metafile
    }
    if ((getline < fname) > 0 && ($0 ~ /^ /))
    {
	sub(/\r/, "")
        sub(/^-+/, "")
	sub(/^ +/, "")
	sub(/ +$/, "")
	$0 = gensub(LINKEXP, "\\2", "g", $0)
	$0 = gensub(WIKIEXPLINK, "\\1", "g", $0)
	if ($0 != "") 
            printf "%s = %s\n", (file ":2"), $0 >> metafile
    }
    close(fname)
    close(metafile)
}

