# parse Wiki format
# Globals: index_anchor, index_level, list, indent_level, blankline, FS, title, last_h2
function parse_wiki(cmd, page, pagefile, revision, comment, pageno)
{
    index_anchor = 0
    index_level = 0
    list["ol"] = 0
    list["ul"] = 0
    indent_level = 0
    blankline = 1
    FS = "[ ]"
    title = -1
    last_h2 = 0

    while (cmd | getline) 
    {
        remove_lr()
        if (parse_code()) continue
        if (parse_index(revision, page, pagefile)) continue
        if (parse_forum(page, pageno)) continue
        if (parse_blankline()) continue
        parse_htmltag()
        if (parse_inlinecode()) continue
        parse_link(page)
        if (parse_heading()) continue
        if (!comment) { if (parse_title()) continue }
        parse_para()
    }
    close(cmd)
    if (last_h2) print "</div>"        
    close_tags()
}

function remove_lr() { sub(/\r$/, "") }

# Globals: embeded, embeded_code, embeded_langauge, $0, r
function parse_code(    cmd)
{
    if ($0 ~ /^<\/(code|pre)>$/)
    {
        print "<div class=\"fragment\">"
        if (embeded_language != "")
        {
            cmd = "enscript --color -E'" embeded_language "' --language=html -o -|" \
                  "awk '/^<HR>/ { p = 0; } /^<PRE>/ { p = 1; } (p == 1) {print;}'"
            print embeded_code | cmd
            close(cmd)
        }
        else
        {
            print "<pre>" embeded_code "</pre>"
        }
        print "</div>"
        embeded = "false"
        embeded_code = ""
        embeded_language = ""
        return 1 # means "next", should skip the rest parsers
    }
    else if ($0 ~ /^<(code|pre)( +lang=[^>]*)?>$/)
    {    
        embeded = "true"
        embeded_code = ""
        if (match($0, "^<code +lang=([^>]*)>$", r))
            embeded_language = clear_str(r[1]);
        else embeded_language = ""
        close_tags()
        return 1 # means "next", should skip the rest parsers
    }
    else if (embeded == "true")
    {
        if (embeded_code == "") embeded_code = $0
        else embeded_code = embeded_code "\n" $0
        return 1 # means "next", should skip the rest parsers
    }
}

#
# generate indices by gathering all headers in the same page
#
function parse_index(revision, page, pagefile,
    cmd)
{
    if ($0 ~ /^<index.*>$/)
    {
        if ($0 !~ /hidden/) 
        {
            # note that these cmd must be different from what is used in parse_wiki
            if (revision) cmd = "co   -q -p'" revision "' " pagefile
            else cmd = "cat  " pagefile
            parse_summary(cmd, page)
        }
        return 1 # means "next", should skip the rest parsers
    }
}

#
# produce a link to comment
# Globals: SCRIPT, DATADIR, CMD_LIST_DIR, LIMIT_COMMENT, last_h2, $0, filelist, r
function parse_forum(page, pageno, 
    order, line, com, path, pagedir, cmd, n, p, edittime, ls_out, pagesize, total,
    startno, endno, i, c)
{
    if ($0 ~ /^<comment( order=[^>]*)?>$/)
    {
        line = $0
        close_tags();
        if (last_h2) {
          print "</div>"
          last_h2 = 0
        }
        com = sprintf(_"COMMENT", SCRIPT "/" page "?post=true");
        if (match(line, "^<comment order=([^>]*)>$", r))
          order = clear_str(r[1])
        path = DATADIR mk_pathname(page)
        pagedir = dirname(path) "/" shorter(filename(path))
        cmd = sprintf(CMD_LIST_DIR, pagedir "/COMMENT")
        n = 0

        while ((cmd | getline ls_out) >0)
        {
            if (match(ls_out, /\t/) > 1) {
                p = substr(ls_out, 1, RSTART - 1)
                edittime = substr(ls_out, RSTART + 1)
                if (match(p, "([^/]+)/(_[^/]+_)$", r) > 0) {
                    filelist[++n] = pagedir "/COMMENT/" p "\t" r[1] "\t" edittime "\t" r[2]
                }
            }
        }
        close(cmd)
        pagesize = LIMIT_COMMENT

        if (pagesize && pageno != "all")
        {
            if (pageno == "") pageno = 1
            total = n / pagesize
            startno = pagesize * (pageno - 1) + 1
            endno = pagesize * pageno
            if (endno > n) endno = n
        } else {
            total = 0
            startno = 1
            endno = n
        }

        if (total > 1)
        {
            com = com "[ "        
            for (i = 1; i < total + 1; i++)
            {
                if (i == pageno) 
                {
                    com = com i " "
                } else {
                    com = com "<a href=\"" SCRIPT "/" page "?postlist=" i \
                          "\">" i " </a>"
                }
            }
            com = com "<a href=\"" SCRIPT "/" page "?postlist=all\">" _"ALL"\
                      "</a> ]"
        }
        com = com "</font>"

        c = 1
        print com "<div id=\"commentmeta\">"
        if (order == "reverse") {
            for (i = n; i >= 1; i--)
            {
                if (c >= startno && c <= endno)
                {
                    split(filelist[i], m, "\t")
                    parse_comment(page, i, m[1], m[2], m[3], m[4])
                }
                c++
            }
        }
        else {
            for (i = 1; i <= n; i++)
            {
                if (c >= startno && c <= endno)
                {
                    split(filelist[i], m, "\t")
                    parse_comment(page, i, m[1], m[2], m[3], m[4])
                }
                c++
            }
        }
        print "</div>"
        return 1 # means "next", should skip the rest parsers
    }
}

# parse each comment
# Globals: DATE_BOX, FMT_DATE, SCRIPT, 
function parse_comment(page, idx, file, posttime, edittime, poster,
                       reply, extra) {
    # skip zero sized comment
    if (system("[ -s " file " ]") != 0) return;
    sub(/^_/, "", poster)
    sub(/_$/, "", poster)
    if (edittime > posttime + 60)
        edittime = sprintf(_"EDITTIME", \
            sprintf(DATE_BOX, strftime(FMT_DATE, edittime)))
    else
        edittime = ""

    reply = sprintf(_"POST_REPLY", SCRIPT "/" page "?post=true&quoteidx=" idx "&quoteid=" posttime ":" poster)
    if (poster == AUTH_USER)
        extra = sprintf(_"EXTRA", SCRIPT "/" page "?post=true&postid=" posttime ":" poster, 
			         SCRIPT "/" page "?postremove=true&postid=" posttime ":" poster)
    else extra = ""

    # anonymous user?
    if (match(poster, /[.]/) > 1) {
        sub("[.]", " from ", poster)
        gsub("_", "", poster)
    } else {
        poster = "<a href=\"" SCRIPT "/" poster "\">" mk_displayname(poster) "</a>"
    }

    printf _"POSTED_BY", idx, poster, \
            sprintf(DATE_BOX, strftime(FMT_DATE, posttime)), \
            reply, extra, edittime
    parse_wiki("cat " file, "", file, "", 1)
}


#
# register blanklines
# this is to help compressing multiple blanklines into one
# Globals: blankline
function parse_blankline()
{
    if ($0 ~ /^ *$/)
    { 
        if (title == -1) title = "";
        if (!blankline) close_tags(); 
        blankline = 1; 
        return 1 # means "next", should skip the rest parsers
    }
}

#
# quote HTML entities for <, > and &
# permitted tags are div|blockquote|code|font|pre|br|a|u|b|i
# Globals: r, $0
function parse_htmltag(    tmp, line, front, tag)
{
    if ($0 ~ /[&<>]/)
    {
        gsub(/&/, "\\&amp;")
        tmp = ""
        line = $0
        while (match(line, "^([^<]*)(<[^>]+>)(.*)$", r))
        {
            if (match(r[1], "^(.*)`$", s))
            {
                front = s[1] r[2]
                tag = ""
                line = r[3]
            }
            else
            {
                front = r[1];
                tag = r[2];
                line = r[3];
                if (!match(tag, "^</?(h[1-6]|div|blockquote|code|font|pre|br|a|u|b|i)( [^>]*)?>$"))
                {
                    front = front tag
                    tag = ""
                }
             }
             gsub(/</, "\\&lt;", front)
             gsub(/>/, "\\&gt;", front)
             tmp = tmp front tag
        }
        gsub(/</, "\\&lt;", line); 
        gsub(/>/, "\\&gt;", line);
        $0 = tmp line
    }
}

# things in $ ... $ are codes
function parse_inlinecode()
{
    if ($0 ~ /.*([^`a-z0-1_]|^)[$]((\w|\/|&).*(\w|;)|\w)[$](\W|$)/)
    {
        $0 = gensub(/([^`a-z0-1_]|^)[$]([^$]+)[$](\W|$)/, "\\1<code>\\2</code>\\3", "g"); 
        gsub(/`[$]/, "$");
    }
}

#
# generate links for WIKIEXP, direct http tag, attachment link,
# and [link | name]
# Globals: WIKIEXPLINK, LINKEXP, r, $0
function parse_link(page,    tmp, line, name, middle, after, deepquote) 
{
    if (($0 ~ WIKIEXPLINK) || ($0 ~ LINKEXP) || /(https?|ftp|gopher|mailto|news):/)
    {
        tmp = ""
        line = $0
        while (match(line, "^([^[]*\\[?)(\\[[^\\[\\]]+\\])(.*)$", r))
        {
            deepquote = sub("\\[$", "", r[1]) && sub("^\\]", "", r[3]);
            tmp = tmp r[1]
            middle = r[2]
            after = r[3]
            if (!match(r[1], /`$/)) 
            {
                # generate [ .. | .. ] style links
                if (match(middle, "^" LINKEXP "$", r))
                {
                    link = r[1]
                    gsub("&", "\\\\&", link)
                    name = r[2]
                    tmp = tmp mk_link(page, link, name)
                    line = after
                    continue;
                }
                else if (match(middle, "^" WIKIEXPLINK "$", r)) 
                {
                    link = r[1]
                    tmp = tmp mk_link(page, link, NULL, deepquote)
                    line = after
                    continue;
                }
                else
                {
                    middle = mk_link(page, middle)
                }
            }
            else if (deepquote) {
                sub(/`$/, "[", tmp)
                middle = middle "]"
            }
            else sub(/`$/, "", tmp)
            tmp = tmp middle
            line = after
        }
        $0 = tmp line
    }
}

#headings
# Globals: last_h2, title, $0, blankline, 
function parse_heading()
{
    if ($0 ~ /^-[^-]/)
    {
        $0 = "<h2>" strip_footnote(substr($0, 2)) "</h2>"; close_tags(); 
        if (title == -1) title = $0
        if (!last_h2)
        {
            last_h2 = 1
            print "<div class='chapter'>"
        }
        else {
            print "</div><div class='chapter'>"
        }
    }
    else if ($0 ~ /^--[^-]/)
    {
        $0 = "<h3>" strip_footnote(substr($0, 3)) "</h3>"; close_tags(); 
        if (title == -1) title = $0
    }
    else if ($0 ~ /^---[^-]/)
    { 
        $0 = "<h4>" strip_footnote(substr($0, 4)) "</h4>"; close_tags(); 
        if (title == -1) title = $0
    }
    else if ($0 ~ /^----/)
    { 
        sub(/^----+/, "<hr>")
        blankline = 1
        close_tags()
        print
        return 1
    }
    else return
    print next_anchor() $0
    return 1
}

# Output and set page title if it's non-empty
# Globals: title, $0
function parse_title()
{
    if (title == -1)
    {
        sub(/^-+/, "")
        sub(/^ +/, "")
        title = $0
        print "<h1>" $0 "</h1>"
        return 1
    }
}

# parse paragraphs and lists
# Globals: r, $0, blanks
function parse_para(    blanks, line)
{ 
    if (match($0, "^( *)([^ ].*)$", r))
    {
        blanks = length(r[1])
        line = r[2]
        if (blanks && match(line, "^1 (.*)$", r))
        {
            match_level(blanks)
            open_tags("ol", "<li>")
            line = r[1]
        }
        else if (blanks && match(line, "^\\* (.*)$", r))
        {
            match_level(blanks)
            open_tags("ul", "<li>")
            line = r[1]
        }
        else if (blanks && match(line, "^# (.*)$", r))
        {
            match_level(blanks)
            open_tags("dl", "<dd>", r[1])
            line = ""
        }
	else if (blanks || blankline)
        {
            matched = match_level(blanks)
            # new tag bl to suppress item heading
	    if (matched) print "<p style='margin: 0px'>"
            else if (blanks) open_tags("any", "<li style='list-style:none;'>")
            else open_tags("p")
        }
        print line
        blankline = 0;
    }
    else print $0;
}

# parse search results
# Globals: r, $0, WIKIEXPFILE
function parse_search(cmd,
    page, name, last_page, t)
{
    print "<table width='100%'>"
    while (cmd | getline) 
    {
        if ($0 ~ WIKIEXPFILE)
        {
           if (match($0, "^(.*):.? = (.*)$", r))
            {
                page = km_pathname(r[1])
                if (page != last_page) {
                    last_page = page
                    name = gensub("_", " ", "g", page)
                    t = strftime(FMT_DATE, get_meta(r[1]))
                    printf _"RECENT_TR", mk_simple_link(page, name), \
                           mk_simple_link(page, get_meta(r[1] ":")), \
                           sprintf(DATE_BOX, t)
                }
            }    
        }
    }
    close(cmd)
    print "</table>"
}

# parse renaming results, and rename old links to new ones
# Globals: WIKIEXPFILE, DATADIR, FILE_META, $0
function parse_rename(cmd0, page, newpage, date, addr,
    cmd, file, tempfile)
{
    while (cmd0 | getline) 
    {
        if ($0 ~ (WIKIEXPFILE "$"))
        {
            cmd = "mktemp /tmp/temp.XXXXXX"
            cmd | getline tempfile
            close(cmd)
            system("sed 's/\\(\\[\\)" page \
                  "\\(\\(:.\\+\\]\\)\\|\\]\\)/\\1" newpage \
                  "\\2/g' '" $1 "' >" tempfile)
            system("chmod go+r " tempfile)
            system("mv " tempfile " '" $1 "'")
            file = $1
            sub ("^" DATADIR, "", file)
            append_meta(FILE_META, systime(), file)
            if (date && addr)
            {
                system("ci -q -t-" $1 " -l -m'" addr ";;" date \
                       ";;affected by rename' " $1)
            }
        }
    }
    close(cmd0)
}

# parse rlog history
# Globals: NR, sp_array, revisions
function parse_history(cmd, page,
    n, author, ip_address, date, comment, revision) 
{
    cmd | getline
    n = 1
    while (cmd | getline) 
    {
        n++
        #pick revision number
        match($1, /[1-9]\.[0-9]+/)
        revision = substr($1, RSTART, RLENGTH)
        match($2, /author: [^;]+;/)
        author = substr($2, RSTART + 8, RLENGTH - 9)
        if (author ~ ("^" WIKIEXP "$")) {
            author = "<a href=\"" SCRIPT "/" author "\">" mk_displayname(author) "</a>"
        } else {
            author = ""
        }
        #split the double-semicolon seperated information line
        split($3, sp_array, ";;")
        ip_address = hide_ip(sp_array[1])
        date = sp_array[2]
        comment = sp_array[3]
        revisions[++num_revisions] = revision
        #if (n == 2) revision = "current"
       
        printf _"HISTORY_TD", revision, sprintf(DATE_BOX, date), ip_address, \
            comment, author, (SCRIPT "/" page (n > 2 ? "?revision=" revision : ""))
    }        
    close(cmd)
}

# parse the index summary
# Globals: $0
function parse_summary(cmd, page,    old_index_anchor)
{
    old_index_anchor = index_anchor
    index_anchor = 0
    while (cmd | getline) 
    {
        parse_htmltag()
        parse_inlinecode()
        parse_link(page)
        if (parse_summary_heading(page)) continue
    }
    close(cmd)
    close_ul(0)
    index_anchor = old_index_anchor
}

# parse heading for summary mode
# Globals: SCRIPT, index_anchor
function parse_summary_heading(page,
    i)
{
    if ($0 ~ /^-[^-]/) i = 1
    else if ($0 ~ /^--[^-]/) i = 2
    else if ($0 ~ /^---[^-]/) i = 3
    else return
    close_ul(i)
    open_ul(i)
    print "<li><a href=\"" SCRIPT "/" page "#anchor_" index_anchor "\">" \
          strip_footnote(substr($0, i + 1)) "</a></li>"
    index_anchor = index_anchor + 1
    return 1
}
