# miscellaneous functions go below
#
# GLOBALS used: _ord_

# convert a wiki name into a display name
function mk_displayname(name)
{
    gsub(/_/, " ", name)
    gsub(/:/, " : ", name)
    return name
}


# convert wiki name to file system path
function mk_pathname(link,
    path, name)
{
    gsub(/ *: */, ":", link)
    gsub(" ", "_", link)
    if (match(link, /^(.*):([^:]+)$/, r))
    {
        path = r[1] ":"
        name = "_" r[2] "_"
        # remove all .. and . in path
        gsub(/\./, "", path)
        # convert path to directory
        gsub(/:/, "/", path)
        link = path name
    }
    else link = "_" link "_"
    return link
}

# get the directory name of a path
function dirname(path,
    r)
{
    if (match(path, /^(.*)\/([^/]+)$/, r))
    {
        return r[1]
    }
    else return ""
}

# get the file name of a path
function filename(path,
   r)
{
    if (match(path, /^(.*)\/([^/]+)$/, r))
    {
        return r[2]
    }
    else return path
}

# go from pathname to full wikiname
function km_pathname(path,
     dir, file)
{
    dir = dirname(path)
    file = filename(path)

    if (dir)
    {
        if (file ~ /^_.*_$/)
        {
           gsub(/\//, ":", dir)
           return dir ":" shorter(file)
        }
        else return ""
    }
    else return shorter(file)
}

function random(n)
{
    return int(rand() * n)
}

function revstr(s,   i, t)
{
    t = ""
    for (i = 1; i <= length(s); i++)
    {
        t = (substr(s, i, 1)) t
    }
    return t
}

function hide_ip(ip)
{
    return gensub("^[^.]*\\.[^.]*\\.(.*)$", "*.*.\\1", 1, ip)
}

function substr_utf8(s, i, n,
    t, j, c0)
{
    j = 0
    t = ""
    while (i <= length(s) && j < n)
    {
        c0 = ord(substr(s, i++, 1))
        if (c0 < 0x7f) t = t substr(s, i-1, 1)
        else if (c0 >= 0xc0 && c0 < 0xe0)
        {
            i++
            t = t substr(s, i-2, 2)
        }
        else if (c0 >= 0xe0)
        {
            i+=2
            t = t substr(s, i-3, 3)
        }
        j++
    }
    return t
}


# this has some legacy code, and could be optimized further
function short_title(title, n, 
    s)
{
    if (!n) n = 30
    s = substr_utf8(title, 1, n)
    if (s != title) return s "..."
    else return title
}

# following codes are for ord and chr function, 
# need _ord_init() in BEGIN
function _ord_init(    low, high, i, t)
{
    low = 0
    high = 255

    for (i = low; i <= high; i++)
    {
        t = sprintf("%c", i)
        _ord_[t] = i
    }
}

function ord(s)
{
    return _ord_[s]
}

function chr(c)
{
    # force c to be numeric by adding 0
    return sprintf("%c", c + 0)
}

function strip_footnote(u,    s)
{
    s = ""
    while (i = index(u, FOOTNOTE_O)) {
       s = s substr(u, 1, i - 1)
       u = substr(u, i + length(FOOTNOTE_O))
       if (s ~ /^ *$/) {
            s = s FOOTNOTE_O
       }
       else {
	   if (i = index(u, FOOTNOTE_C)) {
	      u = substr(u, i + length(FOOTNOTE_C))
	   }
	   else {
	      s = s FOOTNOTE_O u
	      u = ""
	   }
       }
    }
    return s u
}


function sendmail(fromaddr, toaddr, subject, msg)
{
    cmd = sprintf(CMD_SENDMAIL, fromaddr, toaddr, subject)
    print msg | cmd
    #close(cmd)    
}

# get the modification time of the given file
function get_mtime(fname,    ret, cmd)
{
    if (EXTENSION) 
    {
        stat_data[1] = 1
        ret = stat(fname, stat_data)
        if (ret >= 0) return stat_data["mtime"]
    }
    else {
        cmd = sprintf(CMD_LIST_FILE, fname)
        cmd | getline ret
        close(cmd)
        if (match(ret, /(.*)\t(.*)/, r))
        {
            return int(r[2])
        }
    }
}

